"""Validators for Peertube."""

from re import match


def channel_name(value: str) -> bool:
    """Validate a Peertube channel name.

    Args:
        value (str): The proposed channel name.

    Returns:
        bool: Whether the channel name is valid.
    """

    return bool(match(r"^[0-9a-z_.]{1,50}$", value))



"""Generic exceptions for Peertube."""

from requests import Response
from requests.exceptions import JSONDecodeError


class PeertubeError(Exception):
    """A generic Peertube error."""


class PeerTubeAPIBadResponseError(PeertubeError):
    """An unexpected API response was received."""


def raise_api_bad_response_error(response: Response):
    """Raise a PeerTubeAPIBadResponseError for the given response, trying to decode JSON response.

    Args:
        response (Response): The response to raise the error for.
    """

    try:
        raise PeerTubeAPIBadResponseError(
            response.url, response.status_code, response.reason, response.json()
        )
    except JSONDecodeError as e:
        raise PeerTubeAPIBadResponseError(
            response.url,
            response.status_code,
            response.reason,
            response.content.decode(),
        ) from e



"""Tools to upload videos to a PeerTube instance"""

__author__ = "CaramelConnoisseur"
__copyright__ = "Copyright 2025, CaramelConnoisseur"
__email__ = "me@CaramelConnoisseur.dev"
__license__ = "GPL-3.0-only"
__maintainer__ = "CaramelConnoisseur"
__status__ = "Production"
__version__ = "1.1.1"

from dataclasses import dataclass
from datetime import datetime
from typing import Any, Dict, List, Optional


@dataclass
class Image:
    """An image on Peertube."""

    path: str
    width: int
    created_at: datetime
    updated_at: datetime

    def __repr__(self):
        return f"Image: {self.path}"

    def __str__(self):
        return self.path


class Account:  # pylint: disable=too-few-public-methods,too-many-instance-attributes
    """A user account on Peertube."""

    id: int
    url: str
    name: str
    avatars: List[Image]
    host: str
    host_redundancy_allowed: Optional[bool]
    following_count: Optional[int]
    followers_count: Optional[int]
    created_at: Optional[datetime]
    updated_at: Optional[datetime]
    user_id: Optional[int]
    display_name: str
    description: Optional[str]

    def __init__(self, input_dict: Dict[str, Any]):
        self.id = input_dict["id"]
        self.url = input_dict["url"]
        self.name = input_dict["name"]
        self.avatars = []
        for avatar in input_dict["avatars"]:
            self.avatars.append(
                Image(
                    avatar["path"],
                    avatar["width"],
                    avatar["createdAt"],
                    avatar["updatedAt"],
                )
            )
        self.host = input_dict["host"]
        self.host_redundancy_allowed = (
            input_dict["hostRedundancyAllowed"]
            if "hostRedundancyAllowed" in input_dict
            else None
        )
        self.following_count = (
            input_dict["followingCount"] if "followingCount" in input_dict else None
        )
        self.followers_count = (
            input_dict["followersCount"] if "followersCount" in input_dict else None
        )
        self.created_at = input_dict["createdAt"] if "createdAt" in input_dict else None
        self.updated_at = input_dict["updatedAt"] if "updatedAt" in input_dict else None
        self.user_id = input_dict["userId"] if "userId" in input_dict else None
        self.display_name = input_dict["displayName"]
        self.description = (
            input_dict["description"] if "description" in input_dict else None
        )

    def __repr__(self):
        return f"Account: {self.id} - {self.name}"

    def __str__(self):
        return self.name



"""Functions for managing Peertube Video Channels."""

from datetime import datetime
from enum import Enum
from mimetypes import guess_type
from os.path import basename
from typing import Any, Dict, List, Optional, Union

from . import Account, Image, validators
from .client import ApiClient
from .exceptions import PeertubeError, raise_api_bad_response_error


class VideoChannel:  # pylint: disable=too-few-public-methods,too-many-instance-attributes
    """A Video Channel on Peertube."""

    id: int
    url: str
    name: str
    avatars: List[Image]
    host: str
    host_redundancy_allowed: Optional[bool]
    following_count: Optional[int]
    followers_count: Optional[int]
    created_at: Optional[datetime]
    updated_at: Optional[datetime]
    display_name: str
    description: Optional[str]
    support: Optional[str]
    local: Optional[bool]
    banners: Optional[List[Image]]
    owner: Optional[Account]

    def __init__(self, input_dict: Dict[str, Any]):
        self.id = input_dict["id"]
        self.url = input_dict["url"]
        self.name = input_dict["name"]
        self.avatars = []
        for avatar in input_dict["avatars"]:
            self.avatars.append(
                Image(
                    avatar["path"],
                    avatar["width"],
                    avatar["createdAt"],
                    avatar["updatedAt"],
                )
            )
        self.host = input_dict["host"]
        self.host_redundancy_allowed = (
            input_dict["hostRedundancyAllowed"]
            if "hostRedundancyAllowed" in input_dict
            else None
        )
        self.following_count = (
            input_dict["followingCount"] if "followingCount" in input_dict else None
        )
        self.followers_count = (
            input_dict["followersCount"] if "followersCount" in input_dict else None
        )
        self.created_at = (
            datetime.fromisoformat(input_dict["createdAt"])
            if "createdAt" in input_dict
            else None
        )
        self.updated_at = (
            datetime.fromisoformat(input_dict["updatedAt"])
            if "updatedAt" in input_dict
            else None
        )
        self.display_name = input_dict["displayName"]
        self.description = (
            input_dict["description"] if "description" in input_dict else None
        )
        self.support = input_dict["support"] if "support" in input_dict else None
        self.local = input_dict["isLocal"] if "isLocal" in input_dict else None
        if "banners" in input_dict:
            self.banners = []
            for banner in input_dict["banners"]:
                self.banners.append(
                    Image(
                        banner["path"],
                        banner["width"],
                        banner["createdAt"],
                        banner["updatedAt"],
                    )
                )
        else:
            self.banners = None
        self.owner = (
            Account(input_dict["ownerAccount"])
            if "ownerAccount" in input_dict
            else None
        )

    def __repr__(self):
        return f"Channel: {self.id} - {self.name}"

    def __str__(self):
        return self.name


class ChannelEndpoints(Enum):
    """Endpoints for channel management."""

    CREATE_VIDEO_CHANNEL = "api/v1/video-channels"
    """Create a video channel."""

    GET_VIDEO_CHANNELS = "api/v1/video-channels?count=100&sort=name&start={start}"
    """List video channels."""

    VIDEO_CHANNEL = "api/v1/video-channels/{channel}"
    """Get a video channel."""

    SET_VIDEO_CHANNEL_AVATAR = "api/v1/video-channels/{channel}/avatar/pick"
    """Update channel avatar."""

    SET_VIDEO_CHANNEL_BANNER = "api/v1/video-channels/{channel}/banner/pick"
    """Update channel avatar."""


class PeertubeChannelError(PeertubeError):
    """A generic Peertube Channel error."""


class PeertubeNonChannelExistsError(PeertubeChannelError):
    """The requested channel already exists."""


class PeertubeNonExistentChannelError(PeertubeChannelError):
    """The requested channel does not exist."""


def create_channel(  # final call raises exception pylint: disable=inconsistent-return-statements
    client: ApiClient,
    name: str,
    display_name: str,
    description: Optional[str] = None,
    sponsor: Optional[str] = None,
) -> VideoChannel: # pyright: ignore[reportReturnType]
    """Create a new Video Channel on Peertube.

    Args:
        client (ApiClient): The authenticated client to use.
        name (str): The name of the new channel. Up to 50 alphanumeric, `_` or `.` characters.
        display_name (str): _description_
        description (Optional[str]): _description_
        sponsor (Optional[str], optional): _description_. Defaults to None.

    Raises:
        PeertubeNonChannelExistsError: _description_
        PeerTubeAPIBadResponseError: _description_

    Returns:
        int: _description_
    """

    if not validators.channel_name(name):
        raise ValueError("Invalid channel name", name)

    try:
        channel_id = get_channel(client, name)
        raise PeertubeNonChannelExistsError(client.base_url, name, channel_id)
    except PeertubeNonExistentChannelError:
        pass

    response = client.session.post(
        client.base_url + ChannelEndpoints.CREATE_VIDEO_CHANNEL.value,
        json={
            "displayName": display_name,
            "name": name,
            "description": description,
            "sponsor": sponsor,
        },
        timeout=10,
    )
    if response.status_code == 200:
        return get_channel(client, name)

    raise_api_bad_response_error(response)


def delete_channel(client: ApiClient, channel: str):
    """Delete a channel.

    Args:
        client (ApiClient): The authenticated client to use.
        channel (str): The channel to delete.

    Raises:
        ValueError: If the channel name is malformed.
    """

    if not validators.channel_name(channel):
        raise ValueError("Invalid channel name", channel)

    response = client.session.delete(
        client.base_url + ChannelEndpoints.VIDEO_CHANNEL.value.format(channel=channel)
    )

    if response.status_code != 204:
        raise_api_bad_response_error(response)


def get_channel(  # final call raises exception pylint: disable=inconsistent-return-statements
    client: ApiClient, name: str
) -> VideoChannel: # pyright: ignore[reportReturnType]
    """Get the numeric identifier of a given video channel.

    Args:
        client (ApiClient): The authenticated client to use.
        channel_name (str): The channel to look for.

    Raises:
        PeertubeNonExistentChannelError: If the channel does not exist and creation is not enabled.
        PeerTubeAPIBadResponseError: If an error occurs talking to the Peertube API.

    Returns:
        int: The channels ID.
    """

    if not validators.channel_name(name):
        raise ValueError("Invalid channel name", name)

    response = client.session.get(
        client.base_url + ChannelEndpoints.VIDEO_CHANNEL.value.format(channel=name),
        timeout=10,
    )

    if response.status_code == 200:
        return VideoChannel(response.json())

    if response.status_code == 404:
        raise PeertubeNonExistentChannelError(client.base_url, name)

    raise_api_bad_response_error(response)


def get_channels(client: ApiClient) -> List[VideoChannel]:
    """Get list of all know channels.

    Args:
        client (ApiClient): The authenticated client to use.

    Returns:
        List[VideoChannel]: A list of known channels.
    """

    channels: List[VideoChannel] = []
    total = -1
    start = 0
    while len(channels) != total:
        response = client.session.get(
            client.base_url
            + ChannelEndpoints.GET_VIDEO_CHANNELS.value.format(start=start),
            timeout=10,
        )
        response.raise_for_status()
        start = start + 100
        body = response.json()
        total = body["total"]
        for channel in body["data"]:
            channels.append(VideoChannel(channel))

    return channels


def set_channel_avatar(client: ApiClient, channel: str, avatar_path: str):
    """Upload an avatar image to the specified Peertube channel, it will be stretched to a square.

    Args:
        client (ApiClient): The authenticated client to use.
        channel (str): The channel to update.
        avatar_path (str): The path to the file to use as the new avatar.

    Raises:
        ValueError: If the channel name is in the wrong format.
        PeerTubeAPIBadResponseError: If an error occurs talking to the Peertube API.
    """

    if not validators.channel_name(channel):
        raise ValueError("Invalid channel name", channel)

    with open(avatar_path, "br") as f:
        response = client.session.post(
            client.base_url
            + ChannelEndpoints.SET_VIDEO_CHANNEL_AVATAR.value.format(channel=channel),
            files={
                "avatarfile": (
                    basename(avatar_path),
                    f,
                    guess_type(avatar_path)[0],
                )
            }, # pyright: ignore[reportArgumentType]
            timeout=30,
        )
    if not response.status_code == 200:
        raise_api_bad_response_error(response)


def set_channel_banner(client: ApiClient, channel: str, banner_path: str):
    """Upload a banner image to the specified Peertube channel, it will be stretched to 6:1.

    Args:
        client (ApiClient): The authenticated client to use.
        channel (str): The channel to update.
        banner_path (str): The path to the file to use as the new banner.

    Raises:
        ValueError: If the channel name is in the wrong format.
        PeerTubeAPIBadResponseError: If an error occurs talking to the Peertube API.
    """

    if not validators.channel_name(channel):
        raise ValueError("Invalid channel name", channel)

    with open(banner_path, "br") as f:
        response = client.session.post(
            client.base_url
            + ChannelEndpoints.SET_VIDEO_CHANNEL_BANNER.value.format(channel=channel),
            files={
                "bannerfile": (
                    basename(banner_path),
                    f,
                    guess_type(banner_path)[0],
                )
            }, # pyright: ignore[reportArgumentType]
            timeout=30,
        )
    if not response.status_code == 200:
        raise_api_bad_response_error(response)


# pylint: disable=inconsistent-return-statements,too-many-arguments,too-many-positional-arguments
def update_channel(
    client: ApiClient,
    name: str,
    display_name: Optional[str] = None,
    description: Optional[str] = None,
    support: Optional[str] = None,
    update_support_on_videos: bool = False,
) -> VideoChannel: # pyright: ignore[reportReturnType]
    """Create a new Video Channel on Peertube.

    Args:
        client (ApiClient): The authenticated client to use.
        name (str): The name of the new channel. Up to 50 alphanumeric, `_` or `.` characters.
        display_name (str), optional): The new display name. Defaults to None (unchanged).
        description (Optional[str]): The new channel description. Defaults to None (unchanged).
        support (Optional[str], optional): The new support message. Defaults to None (unchanged).
        update_support_on_videos (bool, optional): 
            Whether to update the support message on all channel videos. Defaults to False.

    Raises:
        PeertubeNonChannelExistsError: If the target channel doesn't exist.
        PeerTubeAPIBadResponseError: If an unexpected response is returned by the API.

    Returns:
        VideoChannel: The updated channel.
    """

    if not validators.channel_name(name):
        raise ValueError("Invalid channel name", name)

    get_channel(client, name)

    json: Dict[str, Union[str, bool]] = {}
    if display_name is not None:
        json["displayName"] = display_name
    if description is not None:
        json["description"] = description

    if support is not None:
        json["bulkVideosSupportUpdate"] = update_support_on_videos
        json["support"] = support

    if len(json) > 0:
        response = client.session.put(
            f"{client.base_url}{ChannelEndpoints.VIDEO_CHANNEL.value.format(channel=name)}",
            json=json,
            timeout=10,
        )
        if response.status_code == 204:
            return get_channel(client, name)

        raise_api_bad_response_error(response)



"""Functions for managing playlists on Peertube."""

from contextlib import nullcontext
from datetime import datetime
from enum import Enum
from io import BufferedReader
from os.path import basename
from mimetypes import guess_type
from typing import Any, Dict, Optional, List, Literal, Tuple, Union

from . import Account
from .channels import VideoChannel
from .client import ApiClient
from .exceptions import raise_api_bad_response_error
from .videos import Video


class PlaylistPrivacy(Enum):
    """The privacy setting of a playlist."""

    PUBLIC = 1
    UNLISTED = 2
    PRIVATE = 3


class PlaylistType(Enum):
    """The type of playlist this is."""

    REGULAR = 1
    WATCH_LATER = 2


class PlaylistVideo:
    """A video within a playlist on Peertube."""

    id: int
    position: int
    start_timestamp: Optional[int]
    stop_timestamp: Optional[int]
    video: Video

    def __init__(self, input_dict: Dict[str, Any]):
        self.id = input_dict["id"]
        self.position = input_dict["position"]
        self.start_timestamp = input_dict["startTimestamp"]
        self.stop_timestamp = input_dict["stopTimestamp"]
        self.video = Video(input_dict["video"])

    def __repr__(self):
        return f"Playlist Video: {self.id}"

    def __str__(self):
        return f"{self.id} ({self.video.name})"


class PlaylistEndpoints(Enum):
    """Endpoints for managing playlists on Peertube."""

    PLAYLIST = "api/v1/video-playlists/{playlist_id}"
    """Gets/updates a playlist"""

    PLAYLIST_ITEM = "api/v1/video-playlists/{playlist_id}/videos/{playlist_item_id}"
    """Removes a video from a playlist"""

    PLAYLISTS = "api/v1/video-playlists"
    """Gets all playlists on the instance"""

    PLAYLISTS_IN_ACCOUNT = "api/v1/accounts/{account_name}/video-playlists"
    """Gets list of playlists in a given account"""

    PLAYLISTS_IN_CHANNEL = "api/v1/video-channels/{channel}/video-playlists"
    """Gets list of playlists in a given channel"""

    REORDER_PLAYLISTS_IN_CHANNEL = (
        "api/v1/video-channels/{channel}/video-playlists/reorder"
    )
    """Reorder channel playlists"""

    REORDER_VIDEOS_IN_PLAYLIST = "api/v1/video-playlists/{playlist_id}/videos/reorder"
    """Reorder playlist elements"""

    SEARCH_PLAYLISTS = "api/v1/search/video-playlists"
    """Search for playlists"""

    VIDEOS_IN_PLAYLIST = "api/v1/video-playlists/{playlist_id}/videos"
    """Gets list of videos within a playlist"""


# pylint: disable=too-many-instance-attributes
class Playlist:
    """A playlist of videos on Peertube"""

    id: int
    uuid: str
    short_uuid: str
    channel_position: Optional[int]
    channel: Optional[VideoChannel]
    created_at: datetime
    description: str
    display_name: str
    embed_path: str
    is_local: bool
    owner: Account
    privacy: PlaylistPrivacy
    thumbnail_path: str
    type: PlaylistType
    updated_at: datetime
    url: str
    videos_length: int
    videos: List[PlaylistVideo]

    def __init__(self, client: ApiClient, input_dict: Dict[str, Any]):
        self.id = input_dict["id"]
        self.uuid = input_dict["uuid"]
        self.short_uuid = input_dict["shortUUID"]
        self.channel = (
            VideoChannel(input_dict["videoChannel"])
            if input_dict["videoChannel"] is not None
            else None
        )
        self.channel_position = input_dict["videoChannelPosition"]
        self.created_at = input_dict["createdAt"]
        self.description = input_dict["description"]
        self.display_name = input_dict["displayName"]
        self.embed_path = input_dict["embedPath"]
        self.is_local = input_dict["isLocal"]
        self.owner = Account(input_dict["ownerAccount"])
        self.privacy = PlaylistPrivacy(input_dict["privacy"]["id"])
        self.thumbnail_path = input_dict["thumbnailPath"]
        self.type = PlaylistType(input_dict["type"]["id"])
        self.updated_at = input_dict["updatedAt"]
        self.url = input_dict["url"]
        self.videos_length = input_dict["videosLength"]
        self.videos = get_videos_in_playlist(client, self.id)

    def __repr__(self):
        return f"Playlist: {self.short_uuid}"

    def __str__(self):
        return f"{self.display_name} ({self.short_uuid})"


def add_video_to_playlist(
    client: ApiClient,
    playlist_id: int,
    video_id: Union[int, str],
    *,
    start: Optional[int] = None,
    end: Optional[int] = None,
) -> Literal[True]:
    """Add a video to a playlist.

    Args:
        client (ApiClient): The client to use to talk to the API.
        playlist_id (int): The numeric ID of the playlist to add to.
        video_id (Union[int, str]): The numeric ID, UUID or Short UUID of the video to add.
        start (int, optional): The start second of the sub-clip to add. Defaults to None.
        end (int, optional): The end second of the sub-clip to add. Defaults to None.

    Returns:
        True: If the video is added successfully.
    """

    data = {"videoId": video_id}
    if start is not None:
        data["startTimestamp"] = start
    if end is not None:
        data["stopTimestamp"] = end

    response = client.session.post(
        client.base_url
        + PlaylistEndpoints.VIDEOS_IN_PLAYLIST.value.format(playlist_id=playlist_id),
        data=data,
        timeout=30,
    )

    if response.status_code != 200:
        raise_api_bad_response_error(response)

    return True


# pylint: disable=too-many-arguments,too-many-positional-arguments
def create_playlist(
    client: ApiClient,
    channel_id: int,
    display_name: str,
    description: Optional[str] = None,
    privacy: PlaylistPrivacy = PlaylistPrivacy.PUBLIC,
    thumbnail_file: Optional[str] = None,
) -> Playlist:
    """Create a new playlist.

    Args:
        client (ApiClient): The client to use to talk to the API.
        channel_id (int): The numeric ID of the target channel to create the playlist in.
        display_name (str): The name of the new playlist.
        description (str, optional): A description of the new playlist. Defaults to None.
        privacy (Privacy, optional): The privacy level of the new playlist. Defaults to PUBLIC.
        thumbnail_file (str, optional): The path to a thumbnail for the playlist. Defaults to None.

    Returns:
        Playlist: The newly created playlist.
    """

    data: Dict[str, Union[int, str]] = {
        "videoChannelId": channel_id,
        "displayName": display_name[:120],
        "privacy": privacy.value,
    }
    if description is not None:
        data["description"] = description[:1000]

    files: Dict[str, Tuple[str, BufferedReader, Optional[str]]] = {}

    with (
        open(thumbnail_file, "br") if thumbnail_file is not None else nullcontext()
    ) as thumbnail_f:
        if thumbnail_file is not None:
            files["thumbnailfile"] = (  # type: ignore
                basename(thumbnail_file),
                thumbnail_f,
                guess_type(thumbnail_file)[0],
            )

        response = client.session.post(
            client.base_url + PlaylistEndpoints.PLAYLISTS.value,
            data=data,
            files=files,  # pyright: ignore[reportArgumentType]
            timeout=30,
        )

    if response.status_code != 200:
        raise_api_bad_response_error(response)

    return get_playlist(client, response.json()["videoPlaylist"]["id"])


def delete_playlist(client: ApiClient, playlist_id: int) -> Literal[True]:
    """Delete a playlist.

    Args:
        client (ApiClient): The client to use to talk to the API.
        playlist_id (int): The numeric ID of the playlist.

    Returns:
        True: If the playlist was deleted.
    """

    response = client.session.delete(
        client.base_url
        + PlaylistEndpoints.PLAYLIST.value.format(playlist_id=playlist_id),
        timeout=10,
    )
    if response.status_code != 200:
        raise_api_bad_response_error(response)
    return True


def delete_video_from_playlist(
    client: ApiClient, playlist_id: int, playlist_item_id: int
) -> Literal[True]:
    """Delete a video from a playlist.

    Args:
        client (ApiClient): The client to use to talk to the API.
        playlist_id (int): The numeric ID of the playlist to remove from.
        playlist_item_id (int: The numeric ID of the video to remove.

    Returns:
        True: If the video is removed successfully.
    """

    response = client.session.delete(
        client.base_url
        + PlaylistEndpoints.PLAYLIST_ITEM.value.format(
            playlist_id=playlist_id, playlist_item_id=playlist_item_id
        ),
        timeout=30,
    )

    if response.status_code != 204:
        raise_api_bad_response_error(response)

    return True


def get_playlist(client: ApiClient, playlist_id: int) -> Playlist:
    """Get a playlist.

    Args:
        client (ApiClient): The client to use to talk to the API.
        playlist_id (int): The numeric ID of the playlist.

    Returns:
        Playlist: The requested playlist.
    """

    response = client.session.get(
        client.base_url
        + PlaylistEndpoints.PLAYLIST.value.format(playlist_id=playlist_id),
        timeout=10,
    )
    if response.status_code != 200:
        raise_api_bad_response_error(response)
    return Playlist(client, response.json())


def get_playlists_on_instance(
    client: ApiClient, playlist_type: Optional[PlaylistType] = None
) -> List[Playlist]:
    """Get all playlists in a channel.

    Args:
        client (ApiClient): The client to use to talk to the API.
        playlist_type (PlaylistType, optional): The type of playlist to get. Defaults to None.
        load_videos (bool, optional): Whether to load the list of videos in the playlist.
            Defaults to False.

    Returns:
        List[Playlist]: The playlists in the channel.
    """

    start = 0
    total = -1
    params: Dict[str, Union[str, int]] = {
        "count": 100,
        "sort": "videoChannelPosition",
        "start": start,
    }
    if playlist_type is not None:
        params["type"] = playlist_type.value
    playlists: List[Playlist] = []
    while len(playlists) != total:
        params["start"] = start
        response = client.session.get(
            client.base_url + PlaylistEndpoints.PLAYLISTS.value,
            params=params,
            timeout=10,
        )
        if response.status_code != 200:
            raise_api_bad_response_error(response)
        start = start + 100
        body = response.json()
        total = body["total"]
        for playlist in body["data"]:
            playlists.append(Playlist(client, playlist))
    return playlists


def get_playlists_in_account(
    client: ApiClient, account_name: str, playlist_type: Optional[PlaylistType] = None
) -> List[Playlist]:
    """Get all playlists in an account.

    Args:
        client (ApiClient): The client to use to talk to the API.
        channel (str): The handle of the channel to look in.
        playlist_type (PlaylistType, optional): The type of playlist to get. Defaults to None.
        load_videos (bool, optional): Whether to load the list of videos in the playlist.
            Defaults to False.

    Returns:
        List[Playlist]: The playlists in the channel.
    """

    start = 0
    total = -1
    params: Dict[str, Union[str, int]] = {
        "count": 100,
        "sort": "videoChannelPosition",
        "start": start,
    }
    if playlist_type is not None:
        params["type"] = playlist_type.value
    playlists: List[Playlist] = []
    while len(playlists) != total:
        params["start"] = start
        response = client.session.get(
            client.base_url
            + PlaylistEndpoints.PLAYLISTS_IN_ACCOUNT.value.format(
                account_name=account_name
            ),
            params=params,
            timeout=10,
        )
        if response.status_code != 200:
            raise_api_bad_response_error(response)
        start = start + 100
        body = response.json()
        total = body["total"]
        for playlist in body["data"]:
            playlists.append(Playlist(client, playlist))
    return playlists


def get_playlists_in_channel(
    client: ApiClient, channel: str, playlist_type: Optional[PlaylistType] = None
) -> List[Playlist]:
    """Get all playlists in a channel.

    Args:
        client (ApiClient): The client to use to talk to the API.
        channel (str): The handle of the channel to look in.
        playlist_type (PlaylistType, optional): The type of playlist to get. Defaults to None.
        load_videos (bool, optional): Whether to load the list of videos in the playlist.
            Defaults to False.

    Returns:
        List[Playlist]: The playlists in the channel.
    """

    start = 0
    total = -1
    params: Dict[str, Union[str, int]] = {
        "count": 100,
        "sort": "videoChannelPosition",
        "start": start,
    }
    if playlist_type is not None:
        params["type"] = playlist_type.value
    playlists: List[Playlist] = []
    while len(playlists) != total:
        params["start"] = start
        response = client.session.get(
            client.base_url
            + PlaylistEndpoints.PLAYLISTS_IN_CHANNEL.value.format(channel=channel),
            params=params,
            timeout=10,
        )
        if response.status_code != 200:
            raise_api_bad_response_error(response)
        start = start + 100
        body = response.json()
        total = body["total"]
        for playlist in body["data"]:
            playlists.append(Playlist(client, playlist))
    return playlists


def get_videos_in_playlist(client: ApiClient, playlist_id: int) -> List[PlaylistVideo]:
    """Get a list of videos in a given playlist.

    Args:
        client (ApiClient): The client to use to talk to the API.
        playlist_id (int): The numeric ID of the playlist to load.

    Returns:
        List[Video]: The videos in the playlist.
    """

    start = 0
    total = -1
    videos: List[PlaylistVideo] = []
    while len(videos) != total:
        response = client.session.get(
            client.base_url
            + PlaylistEndpoints.VIDEOS_IN_PLAYLIST.value.format(
                playlist_id=playlist_id
            ),
            params={"count": 100, "start": start},
            timeout=10,
        )
        if response.status_code != 200:
            raise_api_bad_response_error(response)
        start = start + 100
        body = response.json()
        total = body["total"]
        for video in body["data"]:
            if video["video"] is None:
                continue
            videos.append(PlaylistVideo(video))
    return videos


def reorder_channel_playlists(
    client: ApiClient,
    channel: str,
    move_after: int,
    first_item_index: int,
    count: int = 1,
) -> Literal[True]:
    """Reorder channel playlists.

    Args:
        client (ApiClient): The client to use to talk to the API.
        channel (str): The video channel.
        move_after (int): New position for the block to reorder.
        first_item_index (int): Start position of the element to reorder.
        count (int, optional): How many element from first_item_index to reorder. Defaults to 1.

    Returns:
        Literal[True]: If the reorder is successful.
    """

    response = client.session.post(
        client.base_url
        + PlaylistEndpoints.REORDER_PLAYLISTS_IN_CHANNEL.value.format(channel=channel),
        json={
            "insertAfterPosition": move_after,
            "startPosition": first_item_index,
            "reorderLength": count,
        },
        timeout=30,
    )

    if response.status_code != 204:
        raise_api_bad_response_error(response)

    return True


def reorder_playlist_videos(
    client: ApiClient,
    playlist_id: int,
    move_after: int,
    first_item_index: int,
    count: int = 1,
) -> Literal[True]:
    """Reorder playlist videos.

    Args:
        client (ApiClient): The client to use to talk to the API.
        playlist_id (int): The ID of the playlist to reorder.
        move_after (int): New position for the block to reorder.
        first_item_index (int): Start position of the element to reorder.
        count (int, optional): How many element from first_item_index to reorder. Defaults to 1.

    Returns:
        Literal[True]: If the reorder is successful.
    """

    response = client.session.post(
        client.base_url
        + PlaylistEndpoints.REORDER_VIDEOS_IN_PLAYLIST.value.format(
            playlist_id=playlist_id
        ),
        json={
            "insertAfterPosition": move_after,
            "startPosition": first_item_index,
            "reorderLength": count,
        },
        timeout=30,
    )

    if response.status_code != 204:
        raise_api_bad_response_error(response)

    return True


def remove_video_from_playlist(
    client: ApiClient, playlist_id: int, playlist_item_id: int
) -> Literal[True]:
    """Remove a video from a playlist.

    Args:
        client (ApiClient): The client to use to talk to the API.
        playlist_id (int): The numeric ID of the playlist to remove from.
        playlist_item_id (Union[int, str]): The numeric ID of the item in the playlist to remove.

    Returns:
        True: If the video is removed successfully.
    """

    response = client.session.delete(
        client.base_url
        + PlaylistEndpoints.PLAYLIST_ITEM.value.format(
            playlist_id=playlist_id, playlist_item_id=playlist_item_id
        ),
        timeout=30,
    )

    if response.status_code != 200:
        raise_api_bad_response_error(response)

    return True


def search_playlists(
    client: ApiClient, search: str, host: Optional[str] = None
) -> List[Playlist]:
    """Search playlists.

    Args:
        client (ApiClient): The client to use to talk to the API.
        search (str): String to search.
        host (Optional[str], optional): Find elements owned by this host. Defaults to None.

    Returns:
        List[Playlist]: Matching playlists.
    """

    start = 0
    total = -1
    params: Dict[str, Union[int, str]] = {
        "count": 100,
        "sort": "createdAt",
        "start": start,
        "search": search,
    }
    if host is not None:
        params["host"] = host

    playlists: List[Playlist] = []
    while len(playlists) != total:
        params["start"] = start
        response = client.session.get(
            client.base_url + PlaylistEndpoints.SEARCH_PLAYLISTS.value,
            params=params,
            timeout=30,
        )
        if response.status_code != 200:
            raise_api_bad_response_error(response)

        start = start + 100
        body = response.json()
        total = body["total"]
        for playlist in body["data"]:
            playlists.append(Playlist(client, playlist))

    return playlists


# pylint: disable=too-many-arguments
def update_playlist(
    client: ApiClient,
    playlist_id: int,
    *,
    channel_id: Optional[int] = None,
    display_name: Optional[str] = None,
    description: Optional[str] = None,
    privacy: Optional[PlaylistPrivacy] = None,
    thumbnail_file: Optional[str] = None,
) -> Playlist:
    """Update an existing playlist.

    Args:
        client (ApiClient): The client to use to talk to the API.
        playlist_id (int): The numeric ID of the playlist to update.
        channel_id (int, optional): The numeric ID of the channel to move the playlist to.
            Defaults to None.
        display_name (str, optional): The updated display name. Defaults to None.
        description (str, optional): The updated description. Defaults to None.
        privacy (Privacy, optional): The new privacy level for the playlist. Defaults to None.
        thumbnail_file (str, optional): The path to the new thumbnail file. Defaults to None.

    Returns:
        Playlist: The updated playlist.
    """

    playlist = get_playlist(client, playlist_id)

    data: Dict[str, Union[int, str, None]] = {
        "displayName": (
            display_name[:120] if display_name is not None else playlist.display_name
        ),
        "description": (
            description[:1000] if description is not None else playlist.description
        ),
        "privacy": privacy.value if privacy is not None else playlist.privacy.value,
    }

    if channel_id is not None:
        data["videoChannelId"] = channel_id
    elif playlist.channel is not None:
        data["videoChannelId"] = playlist.channel.id
    else:
        data["videoChannelId"] = None

    files: Dict[str, Tuple[str, BufferedReader, Optional[str]]] = {}
    if thumbnail_file is not None:
        with open(thumbnail_file, "br") as thumbnail_f:
            files["thumbnailfile"] = (
                basename(thumbnail_file),
                thumbnail_f,
                guess_type(thumbnail_file)[0],
            )

    response = client.session.put(
        client.base_url
        + PlaylistEndpoints.PLAYLIST.value.format(playlist_id=playlist_id),
        data=data,
        files=files,  # pyright: ignore[reportArgumentType]
        timeout=30,
    )

    if response.status_code != 200:
        raise_api_bad_response_error(response)

    return get_playlist(client, playlist_id)


def update_playlist_video_start_end(
    client: ApiClient, playlist_id: int, playlist_item_id: int, start: int, end: int
) -> Literal[True]:
    """Update a playlist element.

    Args:
        client (ApiClient): The client to use to talk to the API.
        playlist_id (int): The numeric ID of the playlist to update.
        playlist_item_id (int): The numeric ID video in the playlist to update.
        start (int): The start second of the sub-clip to add.
        end (int): The end second of the sub-clip to add.

    Returns:
        Literal[True]: If the update is successful.
    """

    response = client.session.put(
        client.base_url
        + PlaylistEndpoints.PLAYLIST_ITEM.value.format(
            playlist_id=playlist_id, playlist_item_id=playlist_item_id
        ),
        data={"startTimestamp": start, "stopTimestamp": end},
        timeout=30,
    )

    if response.status_code != 204:
        raise_api_bad_response_error(response)

    return True



"""A client to talk to Peertube"""

from enum import Enum
from types import TracebackType
from typing import Optional

from requests import get, post, Session

from . import __version__


class AuthEndpoints(Enum):
    """Endpoints for managing authentication to Peertube."""

    OAUTH_CLIENT = "api/v1/oauth-clients/local"
    """Login prerequisite"""

    LOGIN = "api/v1/users/token"
    """Login"""

    LOGOUT = "api/v1/users/revoke-token"
    """Logout"""


class ApiClient:
    """A client to talk to Peertube"""

    _base_url: str
    _username: str
    _password: str
    _access_token: str
    _session: Session

    def __init__(self, base_url: str, username: str, password: str):

        if not base_url.endswith("/"):
            base_url = base_url + "/"

        oauth_response = get(base_url + AuthEndpoints.OAUTH_CLIENT.value, timeout=10)
        oauth_response.raise_for_status()
        oauth = oauth_response.json()

        auth_response = post(
            base_url + AuthEndpoints.LOGIN.value,
            data={
                "client_id": oauth["client_id"],
                "client_secret": oauth["client_secret"],
                "grant_type": "password",
                "username": username,
                "password": password,
            },
            timeout=10,
        )
        auth_response.raise_for_status()

        self._base_url = base_url
        self._username = username
        self._password = password
        self._access_token = auth_response.json()["access_token"]
        self._session = Session()
        self._session.headers = {
            "Accept": "application/json",
            "Authorization": f"Bearer {self._access_token}",
            "User-Agent": "Mozilla/5.0 +https://github.com/CaramelConnoisseur/peertube-python "
            f"Peertube-Python/{__version__}",
        }

    def __enter__(self) -> "ApiClient":
        return self

    def __exit__(
        self,
        _exception_type: Optional[type[BaseException]],
        _exception_value: Optional[BaseException],
        _traceback: Optional[TracebackType],
    ) -> None:
        self.logout()

    @property
    def base_url(self) -> str:
        """The base URL of the Peertube instance."""
        return self._base_url

    @property
    def username(self) -> str:
        """The authenticated username."""
        return self._username

    @property
    def session(self) -> Session:
        """A requests.Session configured with the required authentication header."""
        return self._session

    def logout(self) -> bool:
        """Revoke an access token.

        Args:
            access_token (str): The access token to revoke.

        Returns:
            bool: Whether the token was invalided.
        """

        self._session.close()
        resp = post(
            self._base_url + AuthEndpoints.LOGOUT.value,
            headers={"Authorization": f"Bearer {self._access_token}"},
            timeout=10,
        )
        return resp.json()["success"]



# SPDX-FileCopyrightText: 2025 CaramelConnoisseur
#
# SPDX-License-Identifier: GPL-3.0-only

"""Setup file for PyPeertube"""

from setuptools import setup

if __name__ == "__main__":
    setup(install_requires=["urllib3>=2.6.3"])



"""Functions for managing videos on Peertube."""

from contextlib import nullcontext
from datetime import datetime
from enum import Enum
from io import BufferedReader
from os.path import basename
from mimetypes import guess_type
from typing import Any, Dict, Optional, List, Tuple, Union

from . import Account
from .channels import VideoChannel
from .client import ApiClient
from .exceptions import raise_api_bad_response_error


class Category(Enum):
    """Video Category"""

    MUSIC = 1
    FILMS = 2
    VEHICLES = 3
    ART = 4
    SPORTS = 5
    TRAVEL = 6
    GAMING = 7
    PEOPLE = 8
    COMEDY = 9
    ENTERTAINMENT = 10
    NEWS_POLITICS = 11
    HOW_TO = 12
    EDUCATION = 13
    ACTIVISM = 14
    SCIENCE_TECH = 15
    ANIMALS = 16
    KIDS = 17
    FOOD = 18


class CommentsPolicy(Enum):
    """Comments Policy"""

    ENABLED = 1
    DISABLED = 2
    REQUIRE_APPROVAL = 3


class Licence(Enum):
    """License applied to the video."""

    BY = 1
    BY_SA = 2
    BY_ND = 3
    BY_NC = 4
    BY_NC_SA = 5
    BY_NC_ND = 6
    PDD = 7


class Privacy(Enum):
    """The privacy setting of a video"""

    PUBLIC = 1
    """Public viewable"""

    UNLISTED = 2
    PRIVATE = 3
    INTERNAL = 4
    PASSWORD = 5


class VideoFile:  # pylint: disable=too-few-public-methods,too-many-instance-attributes
    """A file for a video on Peertube."""

    id: int
    magnet_uri: Optional[str]
    resolution: int
    size: int
    torrent_url: Optional[str]
    torrent_download_url: Optional[str]
    file_url: str
    playlist_url: Optional[str]
    file_download_url: str
    fps: int
    width: int
    height: int
    metadata_url: str
    has_audio: bool
    has_video: bool
    storage: int

    def __init__(self, input_dict: Dict[str, Any]):
        self.id = input_dict["id"]
        self.magnet_uri = input_dict["magnetUri"] if "magnetUri" in input_dict else None
        self.resolution = input_dict["resolution"]["id"]
        self.size = input_dict["size"]
        self.torrent_url = (
            input_dict["torrentUrl"] if "torrentUrl" in input_dict else None
        )
        self.torrent_download_url = (
            input_dict["torrentDownloadUrl"]
            if "torrentDownloadUrl" in input_dict
            else None
        )
        self.file_url = input_dict["fileUrl"]
        self.playlist_url = (
            input_dict["playlistUrl"] if "playlistUrl" in input_dict else None
        )
        self.file_download_url = input_dict["fileDownloadUrl"]
        self.fps = input_dict["fps"]
        self.width = input_dict["width"]
        self.height = input_dict["height"]
        self.metadata_url = input_dict["metadataUrl"]
        self.has_audio = input_dict["hasAudio"]
        self.has_video = input_dict["hasVideo"]
        self.storage = input_dict["storage"]

    def __repr__(self):
        return f"Video File: {self.id} - {self.file_url}"

    def __str__(self):
        return self.file_url


class StreamingPlaylist:  # pylint: disable=too-few-public-methods
    """A HLS playlist for a video on Peertube."""

    id: int
    type: int
    url: str
    segments_sha256_url: str
    files: List[VideoFile]
    redundancies: List[str]

    def __init__(self, input_dict: Dict[str, Any]):
        self.id = input_dict["id"]
        self.type = input_dict["type"]
        self.url = input_dict["playlistUrl"]
        self.segments_sha256_url = input_dict["segmentsSha256Url"]
        self.files = []
        for file in input_dict["files"]:
            self.files.append(VideoFile(file))
        self.redundancies = []
        for redundancy in input_dict["redundancies"]:
            self.redundancies.append(redundancy["baseUrl"])

    def __repr__(self):
        return f"Streaming Playlist: {self.id} - {self.url}"

    def __str__(self):
        return self.url


class VideoScheduledUpdate:  # pylint: disable=too-few-public-methods
    """A scheduled change of privacy for a video on Peertube."""

    update_at: datetime
    new_privacy_level: Privacy

    def __init__(self, input_dict: Dict[str, Union[int, str]]):
        self.update_at = datetime.fromisoformat(str(input_dict["updateAt"]))
        self.new_privacy_level = Privacy(input_dict["privacy"])

    def __repr__(self):
        return f"To {self.new_privacy_level.name} at {self.update_at}"


class VideoState(Enum):
    """The state of a video on Peertube."""

    PUBLISHED = 1
    TO_TRANSCODE = 2
    TO_IMPORT = 3
    WAITING_FOR_LIVE = 4
    LIVE_ENDED = 5
    TO_MOVE_TO_EXTERNAL_STORAGE = 6
    TRANSCODING_FAILED = 7
    TO_MOVE_TO_EXTERNAL_STORAGE_FAILED = 8
    TO_EDIT = 9
    TO_MOVE_TO_FILE_SYSTEM = 10
    TO_MOVE_TO_FILE_SYSTEM_FAILED = 11


class Video:  # pylint: disable=too-few-public-methods,too-many-instance-attributes
    """A video on Peertube."""

    id: int
    uuid: str
    short_uuid: str
    live: bool
    created_at: datetime
    published_at: datetime
    updated_at: datetime
    originally_published_at: datetime
    category: Optional[Category]
    licence: Licence
    language: str
    privacy: Privacy
    truncated_description: str
    duration: int
    aspect_ratio: float
    local: bool
    name: str
    thumbnail_path: str
    preview_path: str
    embed_path: str
    views: int
    likes: int
    dislikes: int
    comments: int
    nsfw: bool
    nsfw_flags: int
    nsfw_summary: Optional[str]
    wait_transcoding: Optional[bool]
    state: Optional[VideoState]
    scheduled_update: Optional[VideoScheduledUpdate]
    blocklisted: Optional[bool]
    blocklist_reason: Optional[str]
    owner: Account
    channel: VideoChannel
    viewers: int
    description: str
    support: Optional[str]
    tags: Optional[List[str]]
    comments_enabled: Optional[bool]
    comments_policy: Optional[CommentsPolicy]
    download_enabled: Optional[bool]
    input_file_updated_at: Optional[datetime]
    tracker_urls: Optional[List[str]]
    files: Optional[List[VideoFile]]
    streaming_playlists: Optional[List[StreamingPlaylist]]

    def __init__(
        self, input_dict: Dict[str, Any]
    ):  # pylint: disable=too-many-statements
        self.id = input_dict["id"]
        self.uuid = input_dict["uuid"]
        self.short_uuid = input_dict["shortUUID"]
        self.live = input_dict["isLive"]
        self.created_at = input_dict["createdAt"]
        self.published_at = input_dict["publishedAt"]
        self.updated_at = input_dict["updatedAt"]
        self.originally_published_at = input_dict["originallyPublishedAt"]
        self.category = (
            Category(input_dict["category"]["id"])
            if input_dict["category"]["id"] is not None
            else None
        )
        self.licence = Licence(input_dict["licence"]["id"])
        self.language = input_dict["language"]["id"]
        self.privacy = Privacy(input_dict["privacy"]["id"])
        self.truncated_description = input_dict["truncatedDescription"]
        self.duration = input_dict["duration"]
        self.aspect_ratio = input_dict["aspectRatio"]
        self.local = input_dict["isLocal"]
        self.name = input_dict["name"]
        self.thumbnail_path = input_dict["thumbnailPath"]
        self.preview_path = input_dict["previewPath"]
        self.embed_path = input_dict["embedPath"]
        self.views = input_dict["views"]
        self.likes = input_dict["likes"]
        self.dislikes = input_dict["dislikes"]
        self.comments = input_dict["comments"] if "comments" in input_dict else 0
        self.nsfw = input_dict["nsfw"]
        self.nsfw_flags = input_dict["nsfwFlags"] if "nsfwFlags" in input_dict else 0
        self.nsfw_summary = (
            input_dict["nsfwSummary"] if "nsfwSummary" in input_dict else None
        )
        self.wait_transcoding = (
            input_dict["waitTranscoding"] if "waitTranscoding" in input_dict else None
        )
        self.state = (
            VideoState(input_dict["state"]["id"]) if "state" in input_dict else None
        )
        self.scheduled_update = (
            VideoScheduledUpdate(input_dict["scheduledUpdate"])
            if "scheduledUpdate" in input_dict
            else None
        )
        self.blocklisted = (
            input_dict["blacklisted"] if "blacklisted" in input_dict else None
        )
        self.blocklist_reason = (
            input_dict["blacklistedReason"]
            if "blacklistedReason" in input_dict
            else None
        )
        self.owner = Account(input_dict["account"])
        self.channel = VideoChannel(input_dict["channel"])
        self.viewers = input_dict["viewers"]
        self.description = input_dict["description"]
        self.support = input_dict["support"] if "support" in input_dict else None
        self.tags = input_dict["tags"] if "tags" in input_dict else None
        self.comments_enabled = (
            input_dict["commentsEnabled"] if "commentsEnabled" in input_dict else None
        )
        self.comments_policy = (
            CommentsPolicy(input_dict["commentsPolicy"]["id"])
            if "commentsPolicy" in input_dict
            else None
        )
        self.download_enabled = (
            input_dict["downloadEnabled"] if "downloadEnabled" in input_dict else None
        )
        self.input_file_updated_at = (
            input_dict["inputFileUpdatedAt"]
            if "inputFileUpdatedAt" in input_dict
            else None
        )
        self.tracker_urls = (
            input_dict["trackerUrls"] if "trackerUrls" in input_dict else None
        )
        if "files" in input_dict:
            self.files = []
            for file in input_dict["files"]:
                self.files.append(VideoFile(file))
        else:
            self.files = None
        if "streamingPlaylists" in input_dict:
            self.streaming_playlists = []
            for playlist in input_dict["streamingPlaylists"]:
                self.streaming_playlists.append(StreamingPlaylist(playlist))
        else:
            self.streaming_playlists = None

    def __repr__(self):
        return f"Video: {self.uuid}"

    def __str__(self):
        return f"{self.name} ({self.uuid})"


class VideoEndpoints(Enum):
    """Endpoints for managing videos on Peertube."""

    VIDEO = "api/v1/videos/{id}"
    """Get a video"""

    UPLOAD_VIDEO = "api/v1/videos/upload"
    """Upload a new video"""

    VIDEO_SEARCH = "api/v1/search/videos"
    """Search for videos"""

    VIDEOS_IN_ACCOUNT = "api/v1/accounts/{account}/videos"
    """Gets list of videos in a given account"""

    VIDEOS_IN_CHANNEL = "api/v1/video-channels/{channel}/videos"
    """Gets list of videos in a given channel"""


def delete_video(client: ApiClient, video_id: str):
    """Delete a video.

    Args:
        client (ApiClient): The authenticated client to use.
        short_uuid (str): The ShortUUID of the video to delete.
    """

    response = client.session.delete(
        client.base_url + VideoEndpoints.VIDEO.value.format(id=video_id),
        timeout=10,
    )
    if response.status_code != 204:
        raise_api_bad_response_error(response)


def get_video(  # pylint: disable=inconsistent-return-statements
    client: ApiClient, video_id: str
) -> Video:  # pyright: ignore[reportReturnType]
    """Get details about a video on Peertube.

    Args:
        client (ApiClient): The authenticated client to use.
        video_id (str): The id, UUID, or ShortUUID of the video.

    Returns:
        Video: The details of the video.
    """

    response = client.session.get(
        client.base_url + VideoEndpoints.VIDEO.value.format(id=video_id),
        timeout=10,
    )
    if response.status_code == 200:
        return Video(response.json())

    raise_api_bad_response_error(response)

def get_videos_in_account(client: ApiClient, account: str) -> List[Video]:
    """Get a list of videos in a account.

    Args:
        client (ApiClient): The authenticated client to use.
        account (str): The name of the account to look in.

    Returns:
        List[Video]: The videos in the specified account.
    """

    start = 0
    total = -1
    videos: List[Video] = []
    while len(videos) != total:
        response = client.session.get(
            client.base_url
            + VideoEndpoints.VIDEOS_IN_ACCOUNT.value.format(account=account),
            params={"count": 100, "start": start},
            timeout=10,
        )
        if response.status_code != 200:
            raise_api_bad_response_error(response)
        start = start + 100
        body = response.json()
        total = body["total"]
        for video in body["data"]:
            videos.append(Video(video))
    return videos


def get_videos_in_channel(client: ApiClient, channel: str) -> List[Video]:
    """Get a list of videos in a channel.

    Args:
        client (ApiClient): The authenticated client to use.
        channel (str): The name of the channel to look in.

    Returns:
        List[Video]: The videos in the specified channel.
    """

    start = 0
    total = -1
    videos: List[Video] = []
    while len(videos) != total:
        response = client.session.get(
            client.base_url
            + VideoEndpoints.VIDEOS_IN_CHANNEL.value.format(channel=channel),
            params={"count": 100, "start": start},
            timeout=10,
        )
        if response.status_code != 200:
            raise_api_bad_response_error(response)
        start = start + 100
        body = response.json()
        total = body["total"]
        for video in body["data"]:
            videos.append(Video(video))
    return videos


def search_videos(  # pylint: disable=:too-many-arguments,too-many-locals,too-many-branches
    client: ApiClient,
    search: str,
    *,
    category: Optional[Union[Category, List[Category]]] = None,
    duration_max: Optional[int] = None,
    duration_min: Optional[int] = None,
    include_nsfw: Optional[bool] = None,
    live: Optional[bool] = None,
    published_before: Optional[datetime] = None,
    published_after: Optional[datetime] = None,
    tags_and: Optional[Union[str, List[str]]] = None,
    tags_or: Optional[Union[str, List[str]]] = None,
) -> List[Video]:
    """Search for videos on Peertube.

    Args:
        client (ApiClient): The authenticated client to use.
        search (str): The search string.
        category (Category | List[Category], optional): The category or categories to look in.
            Defaults to None.
        duration_max (int, optional): The maximum duration in seconds. Defaults to None.
        duration_min (int, optional): The minimum duration in seconds. Defaults to None.
        include_nsfw (bool, optional): Whether to include NSFW content. Defaults to None.
        live (bool, optional): Limit to Live or VOD videos. Defaults to None.
        published_before (datetime, optional): Limit to videos published before date time.
            Defaults to None.
        published_after (datetime, optional): Limit to video published after date time.
            Defaults to None.
        tags_and (str | List[str], optional): Limit to videos with all these tags.
            Defaults to None.
        tags_or (str | List[str], optional): Limit to videos with any of these tags.
            Defaults to None.

    Returns:
        List[Video]: The matching videos.
    """

    search_parameters: Dict[str, Any] = {
        "count": 100,
        "search": search,
        "sort": "publishedAt",
        "start": 0,
    }
    if category is not None:
        if isinstance(category, Category):
            search_parameters["categoryOneOf"] = category.value
        else:
            search_parameters["categoryOneOf"] = []
            for cat in category:
                search_parameters[
                    "categoryOneOf"
                ].append(  # pyright: ignore[reportUnknownMemberType]
                    cat.value
                )
    if duration_max is not None:
        search_parameters["durationMax"] = duration_max
    if duration_min is not None:
        search_parameters["durationMin"] = duration_min
    if include_nsfw is not None:
        search_parameters["nsfw"] = include_nsfw
    if live is not None:
        search_parameters["isLive"] = live
    if published_after is not None:
        search_parameters["startDate"] = published_after.isoformat()
    if published_before is not None:
        search_parameters["endDate"] = published_before.isoformat()
    if tags_and is not None:
        search_parameters["tagsAllOf"] = tags_and
    if tags_or is not None:
        search_parameters["tagsOneOf"] = tags_or

    total = -1
    matches: List[Video] = []
    while len(matches) != total:
        response = client.session.get(
            client.base_url + VideoEndpoints.VIDEO_SEARCH.value,
            params=search_parameters,
            timeout=10,
        )
        if response.status_code == 200:
            body = response.json()
            total = body["total"]
            for video in body["data"]:
                matches.append(Video(video))
        else:
            raise_api_bad_response_error(response)

        search_parameters["start"] = search_parameters["start"] + 100

    return matches


# pylint: disable=inconsistent-return-statements,too-many-arguments,too-many-branches,too-many-locals
def upload_video(
    client: ApiClient,
    channel_id: int,
    name: str,
    video_file: str,
    *,
    category: Optional[Category] = None,
    comments_policy: Optional[CommentsPolicy] = None,
    description: Optional[str] = None,
    download_enabled: Optional[bool] = None,
    generate_transcription: Optional[bool] = None,
    language: Optional[str] = None,
    licence: Optional[Licence] = None,
    nsfw: Optional[bool] = None,
    originally_published: Optional[datetime] = None,
    preview_file: Optional[str] = None,
    privacy: Optional[Privacy] = None,
    support: Optional[str] = None,
    tags: Optional[List[str]] = None,
    thumbnail_file: Optional[str] = None,
    video_passwords: Optional[List[str]] = None,
    wait_transcoding: Optional[bool] = None,
) -> Video:  # pyright: ignore[reportReturnType]
    """Upload a video to Peertube.

    Args:
        too (ApiClient): The authenticated client to use.
        too (_type_): _description_
        channel_id (int): The numeric identifier of the channel to upload to.
        name (str): The title for the new video.
        video_file (str): The path to the video to upload.
        category (Category, optional): The category to associated to the video. Defaults to None.
        comments_policy (CommentsPolicy, optional): The comments policy to apply to the video.
            Defaults to None.
        description (str, optional): The description to set on the video. Defaults to None.
        download_enabled (bool, optional): Whether to allow downloading of the video.
            Defaults to None.
        generate_transcription (bool, optional): Whether to generate subtitles for the video.
            Defaults to None.
        language (str, optional): The two letter language code for the video. Defaults to None.
        licence (Licence, optional): The license to apply to the video. Defaults to None.
        nsfw (bool, optional): Whether the video contains Not Safe For Work (NSFW) themes.
            Defaults to None.
        originally_published (datetime, optional): When the video was originally published.
            Defaults to None.
        preview_file (str, optional): The image to show on the video page before playing,
            will be stretched to the video aspect ratio. Defaults to None.
        privacy (Privacy, optional): The privacy level to apply to the video. Defaults to None.
        support (str, optional): The support message to associate with the video. Defaults to None.
        tags (List[str], optional): The tags to apply to the video. Defaults to None.
        thumbnail_file (str, optional): The image to show on channel listings and search results,
            will be stretched to a 9:5 aspect ratio. Defaults to None.
        video_passwords (List[str], optional): The passwords to accept for the video.
            Defaults to None.
        wait_transcoding (bool, optional): Whether to wait for transcoding to finish before
            publishing the video. Defaults to None.

    Returns:
        Video: The uploaded video.
    """

    metadata: Dict[str, Union[str, int, List[str]]] = {
        "channelId": channel_id,
        "name": name[:120],
    }
    if category is not None:
        metadata["category"] = category.value
    if comments_policy is not None:
        metadata["commentsPolicy"] = comments_policy.value
    if description is not None:
        metadata["description"] = description
    if download_enabled is not None:
        metadata["downloadEnabled"] = download_enabled
    if generate_transcription is not None:
        metadata["generateTranscription"] = generate_transcription
    if language is not None:
        metadata["language"] = language
    if licence is not None:
        metadata["licence"] = licence.value
    if nsfw is not None:
        metadata["nsfw"] = nsfw
    if originally_published is not None:
        metadata["originallyPublishedAt"] = originally_published.strftime(
            "%Y-%m-%d %H:%M"
        )
    if privacy is not None:
        metadata["privacy"] = privacy.value
    if support is not None:
        metadata["support"] = support
    if tags is not None:
        metadata["tags"] = tags
    if video_passwords is not None:
        metadata["videoPasswords"] = video_passwords
    if wait_transcoding is not None:
        metadata["waitTranscoding"] = wait_transcoding

    with open(video_file, "br") as video_f, (
        open(preview_file, "br") if preview_file is not None else nullcontext()
    ) as preview_f, (
        open(thumbnail_file, "br") if thumbnail_file is not None else nullcontext()
    ) as thumbnail_f:
        files: Dict[str, Tuple[str, BufferedReader, Optional[str]]] = {
            "videofile": (
                basename(video_file),
                video_f,
                guess_type(video_file)[0],
            )
        }
        if preview_file is not None and preview_f is not None:
            files["previewfile"] = (
                basename(preview_file),
                preview_f,
                guess_type(preview_file)[0],
            )
        if thumbnail_file is not None and thumbnail_f is not None:
            files["thumbnailfile"] = (
                basename(thumbnail_file),
                thumbnail_f,
                guess_type(thumbnail_file)[0],
            )

        response = client.session.post(
            client.base_url + VideoEndpoints.UPLOAD_VIDEO.value,
            data=metadata,
            files=files,  # pyright: ignore[reportArgumentType]
            timeout=900,
        )
    if response.status_code == 200:
        return get_video(client, response.json()["video"]["id"])

    raise_api_bad_response_error(response)


# pylint: disable=too-many-arguments,too-many-locals,too-many-branches,inconsistent-return-statements
def update_video(
    client: ApiClient,
    video_id: str,
    *,
    category: Optional[Category],
    comments_policy: Optional[CommentsPolicy],
    description: Optional[str],
    download_enabled: Optional[bool],
    language: Optional[str],
    licence: Optional[Licence],
    name: Optional[str],
    nsfw: Optional[bool],
    nsfw_flags: Optional[int],
    nsfw_summary: Optional[str],
    originally_published_at: Optional[datetime],
    preview_file: Optional[str],
    privacy: Optional[Privacy],
    support: Optional[str],
    tags: Optional[List[str]],
    thumbnail_file: Optional[str],
    video_passwords: Optional[List[str]],
    wait_transcoding: Optional[bool],
) -> Optional[Video]:
    """Not yet implemented"""

    update: Dict[str, Union[bool, int, str, List[str]]] = {}
    if category is not None:
        update["category"] = category.value
    if comments_policy is not None:
        update["commentsPolicy"] = comments_policy.value
    if description is not None:
        update["description"] = description
    if download_enabled is not None:
        update["downloadEnabled"] = download_enabled
    if language is not None:
        update["language"] = language
    if licence is not None:
        update["licence"] = licence.value
    if name is not None:
        update["name"] = name
    if nsfw is not None:
        update["nsfw"] = nsfw
    if nsfw_flags is not None:
        update["nsfwFlags"] = nsfw_flags
    if nsfw_summary is not None:
        update["nsfwSummary"] = nsfw_summary
    if originally_published_at is not None:
        update["originallyPublishedAt"] = originally_published_at.strftime(
            "%Y-%m-%d %H:%M"
        )
    if privacy is not None:
        update["privacy"] = privacy.value
    if support is not None:
        update["support"] = support
    if tags is not None:
        update["tags"] = tags
    if video_passwords is not None:
        update["videoPasswords"] = video_passwords
    if wait_transcoding is not None:
        update["waitTranscoding"] = wait_transcoding

    with (
        open(preview_file, "br") if preview_file is not None else nullcontext()
    ) as preview_f, (
        open(thumbnail_file, "br") if thumbnail_file is not None else nullcontext()
    ) as thumbnail_f:
        files: Dict[str, Tuple[str, BufferedReader, Optional[str]]] = {}
        if preview_file is not None and preview_f is not None:
            files["previewfile"] = (
                basename(preview_file),
                preview_f,
                guess_type(preview_file)[0],
            )
        if thumbnail_file is not None and thumbnail_f is not None:
            files["thumbnailfile"] = (
                basename(thumbnail_file),
                thumbnail_f,
                guess_type(thumbnail_file)[0],
            )

        if len(update) == 0 and len(files) == 0:
            return
        response = client.session.put(
            client.base_url + VideoEndpoints.VIDEO.value.format(id=video_id),
            data=update,
            files=files, # pyright: ignore[reportArgumentType]
            timeout=300,
        )

    if response.status_code == 204:
        return get_video(client, video_id)

    raise_api_bad_response_error(response)
